"use strict";
/*
 * Copyright (c) 2018, salesforce.com, inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.fetchAndDisplayJobStatus = exports.fetchAndDisplayBatchStatus = exports.DataBulkStatusCommand = void 0;
const Display = require("../force-cli/force-cli-display");
const Config = require("../force-cli/force-cli-config");
const Messages = require("../force-cli/force-cli-messages");
class DataBulkStatusCommand {
    async execute(context) {
        context.ux.startSpinner('Getting Status');
        let conn = await Config.getActiveConnection(context);
        if (context.flags.jobid && context.flags.batchid) {
            // view batch status
            const batchStatus = await exports.fetchAndDisplayBatchStatus(conn, context.flags.jobid, context.flags.batchid);
            context.ux.stopSpinner();
            return batchStatus;
        }
        else {
            // view job status
            const jobStatus = await exports.fetchAndDisplayJobStatus(conn, context.flags.jobid);
            context.ux.stopSpinner();
            return jobStatus;
        }
    }
}
exports.DataBulkStatusCommand = DataBulkStatusCommand;
/**
 * get and display the batch status
 * exposed for unit testing
 * @param conn {Connection}
 * @param jobId {string}
 * @param batchId {string}
 */
exports.fetchAndDisplayBatchStatus = async function (conn, jobId, batchId) {
    let job = await conn.bulk.job(jobId);
    let found = false;
    let batches = await job.list();
    batches.forEach(function (batch) {
        if (batch.id === batchId) {
            Display.bulkBatchStatus(batch);
            found = true;
        }
    });
    if (!found) {
        throw new Error(Messages.get('DataBulkStatusNoBatchFound', batchId, jobId));
    }
    return batches;
};
/**
 * get and display the job status; close the job if completed
 * @param conn {Connection}
 * @param jobId {string}
 */
exports.fetchAndDisplayJobStatus = async function (conn, jobId, doneCallback) {
    let job = conn.bulk.job(jobId);
    let jobInfo = await job.check();
    Display.bulkJobStatus(jobInfo);
    if (doneCallback) {
        doneCallback({ job: jobInfo });
    }
    return jobInfo;
};

//# sourceMappingURL=dataBulkStatusCommand.js.map
