"use strict";
/*
 * Copyright (c) 2018, salesforce.com, inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
 */
Object.defineProperty(exports, "__esModule", { value: true });
const path = require("path");
const core_1 = require("@salesforce/core");
const _ = require("lodash");
const BBPromise = require("bluebird");
const fs = BBPromise.promisifyAll(require('fs'));
const utils = require("../../core/srcDevUtil");
const Messages = require("../../messages");
const messages = Messages();
const UpgradeAction_1 = require("./UpgradeAction");
const heads_down_project_1 = require("./heads-down-project");
const heads_down_project_2 = require("./heads-down-project");
const DEFAULT = path.join('config', '*def.json');
async function default_1(projectDir, prompt) {
    let answer = await prompt(messages.getMessage('prompt_orgDefPattern', [DEFAULT], 'projectUpgrade'));
    if (answer === 'SKIP') {
        // No action
        return BBPromise.resolve(null);
    }
    if (answer === 'D' || answer === 'DEFAULT') {
        answer = DEFAULT;
    }
    answer = answer.replace(/[.?+^$[\]\\(){}|-]/g, '\\$&');
    answer = answer.replace('*', '.*');
    const fileRegex = new RegExp(`${answer}$`);
    // Find file names that match the regex
    const matchedFiles = [];
    await core_1.fs.actOn(projectDir, async (file) => {
        if (file.match(fileRegex)) {
            matchedFiles.push(file);
        }
    });
    // Check matched files for heads up
    const files = [];
    for (let matchedFile of matchedFiles) {
        // Don't store this so we don't put a bunch of files in memory. The write is going to happen at a later time.
        const contents = await fs.readFileAsync(matchedFile, 'utf8');
        // Check headsUp, company, country, and orgPreference format.
        const needsAction = contents.match(heads_down_project_1.HEADS_UP) ||
            contents.match(/"[cC]ompany"\s*:/) ||
            contents.match(/"[eE]mail"\s*:/) ||
            contents.match(/"[oO]rgPreferences"\s*:\s*\{\s*"(?!(enabled|disabled))/) ||
            contents.match(/"[lL]astName"\s*:/);
        if (needsAction) {
            files.push(matchedFile);
        }
    }
    if (files.length > 0) {
        return BBPromise.resolve(new UpgradeAction_1.default(messages.getMessage('action_orgDefConversion', [files.length], 'projectUpgrade'), async () => {
            // Not doing this in parrellel right now, but we could
            for (let file of files) {
                let json = await utils.readJSON(file);
                // Org preferences need to stay in upper, so  move to array form first
                const orgPreferences = json.orgPreferences || json.OrgPreferences;
                if (_.isPlainObject(orgPreferences) && !orgPreferences.enabled && !orgPreferences.disabled) {
                    const enabled = _.map(orgPreferences, (enabled, prefName) => (!!enabled ? prefName : null)).filter(_.isString);
                    const disabled = _.map(orgPreferences, (enabled, prefName) => (!enabled ? prefName : null)).filter(_.isString);
                    delete json.OrgPreferences;
                    json.orgPreferences = { enabled, disabled };
                }
                const contents = JSON.stringify(json).replace(heads_down_project_1.HEADS_UP, heads_down_project_2.REPLACE_FN);
                json = await utils.parseJSON(contents, file);
                if (json.company) {
                    json.orgName = json.company;
                    delete json.company;
                }
                if (json.email) {
                    json.adminEmail = json.email;
                    delete json.email;
                }
                if (json.lastName) {
                    delete json.lastName;
                }
                await fs.writeFileAsync(file, JSON.stringify(json, null, 2));
            }
            return BBPromise.resolve(null);
        }));
    }
    return BBPromise.resolve(null);
}
exports.default = default_1;

//# sourceMappingURL=org-def-conversion.js.map
